#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdarg.h>
#include <errno.h>
#include <libgen.h>
#include <pthread.h>
#include <sys/ipc.h>
#include <sys/sem.h>

#include <adwin-macros.h>
#include "linux/adwin.h"
#include "libadwin.h"
#include "libadwin-macros.h"
#include "libadwin/adctl.h"
#include "libadwin/limits.h"
#include "libadwin/bits.h"
#include "libadwin/errno.h"
#include "libadwin/types.h"
#include "configfile.h"
#include "libadwin_undocumented.h"
#include "libadwin.h"
#include "linux/adwin.h"

#ifdef HAVE_SYS_UTSNAME_H
#include <sys/utsname.h>
#endif

#ifdef WITH_ETHERNET
#  include <sys/un.h>
#  include <netinet/in.h>
#  include <arpa/inet.h>
#  include <sys/time.h>
#  define USE_FOR_LIBADWIN 1
#  include "ethernet.h"	/* Generischer Treiber für Ethernet */ //todo
#  include "ADEthernet.h"	/* Generischer Treiber für Ethernet */
#  define Generic_Driver_Get_Par_All Generic_Driver_Get_Par_Block
#  define Generic_Driver_Get_FPar_All Generic_Driver_Get_FPar_Block
#endif
#ifdef WITH_LINK
#  include "transputer.h" 	/* Prototypen für Transputer-Funktionen (T_*) */
#  include "dsp.h" 		/* Prototypen für DSP-Funktionen (DSP_*) */
#endif
#ifdef WITH_USB
#  include "usb.h" 		/* Prototypen für USB-Funktionen (USB_*) */
#endif

#ifdef __FreeBSD__
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/sem.h>
#endif


/* Thread-specific errno (begin) */
static pthread_key_t adwin_errno_key;
static pthread_once_t adwin_errno_once = PTHREAD_ONCE_INIT;

char VersionString[255];

static void __adwin_errno_destructor(void *ptr) {
	free(ptr);
}

static void __adwin_errno_once(void) {
	pthread_key_create(&adwin_errno_key, __adwin_errno_destructor);
}

inline int *__adwin_errno_location(void) {
	int *p;
	pthread_once(&adwin_errno_once, __adwin_errno_once);

	if ((p = pthread_getspecific(adwin_errno_key)) == NULL) {
		p = malloc(sizeof(int));
		*p = 0;
		pthread_setspecific(adwin_errno_key, p);
	}

	return p;
}
/* Thread-specific errno (end) */

/* Thread-specific device_no */
static pthread_key_t adwin_deviceno_key;
static pthread_once_t adwin_deviceno_once = PTHREAD_ONCE_INIT;

static void __adwin_deviceno_destructor(void *ptr) {
	free(ptr);
}

static void __adwin_deviceno_once(void) {
	pthread_key_create(&adwin_deviceno_key, __adwin_deviceno_destructor);
}

inline DeviceNumber *__adwin_deviceno_location(void) {
	DeviceNumber *p;
	pthread_once(&adwin_deviceno_once, __adwin_deviceno_once);

	if ((p = pthread_getspecific(adwin_deviceno_key)) == NULL) {
		p = (DeviceNumber*) malloc (sizeof (DeviceNumber));
		p->DeviceType = 0;			// Link
		p->DeviceNo = 0;			// Nicht gesetzt
		pthread_setspecific(adwin_deviceno_key, p);
	}

	return p;
}
/* Thread-spezific device_no (end) */


/***********************************************************************
 * Semaphores 
 */
// global variables for semaphore locking

int semID_glob, semID_dev;

static inline void init_glob_semaphore() {
	semID_glob = semget(SEM_KEY + 1, 1, 0666 | IPC_CREAT | IPC_EXCL);
	
	if (semID_glob >= 0) { 
		// it worked -- we are the first library instance
		PRINTDEBUG(DBG_SEMAPHORES, 
			   ("global semaphore created at %d\n", semID_glob));
		// initialize semaphore 
		union semun sem_init;
		unsigned short values[1];
		values[0] = 1; 
		sem_init.array = values; 
		semctl(semID_glob, 0, SETALL, sem_init);
	} else if (errno==EEXIST) {
		// the semaphore set already exists, so just use it
		if ((semID_glob = semget(SEM_KEY + 1, 1, 0666)) < 0) {
			perror("semget global");
		}
		PRINTDEBUG(DBG_SEMAPHORES, 
			   ("using existing global semaphore %d\n",
			    semID_glob));
	} else {
		perror("semget global");
	}
}

static inline void grab_glob_semaphore() {
	struct sembuf sem = {0, -1, SEM_UNDO};
	
	if (semop(semID_glob, &sem, 1) == -1) {
		perror("grab global semaphore");
	}
	PRINTDEBUG(DBG_SEMAPHORES, ("Grabbed global semaphore, value is %d\n", 
				    semctl(semID_glob, 0, GETVAL, NULL)));
}

static inline void release_glob_semaphore() {
	struct sembuf sem = {0, 1, SEM_UNDO};
	
	if (semop(semID_glob, &sem, 1) == -1) {
		perror("release global semaphore");
	}
	PRINTDEBUG(DBG_SEMAPHORES, 
		   ("Released global semaphore, value is %d\n", 
		    semctl(semID_glob, 0, GETVAL, NULL)));
}

static inline void init_dev_semaphore() {
	// try to create new semaphore set
	semID_dev = semget(SEM_KEY, MAX_DEVICES, 0666 | IPC_CREAT | IPC_EXCL);
	if (semID_dev >= 0) { 
		// it worked -- we are the first library instance
		PRINTDEBUG(DBG_SEMAPHORES, 
			   ("device semaphores created at %d\n", semID_dev));
		// initialize semaphore set
		union semun sem_init;
		unsigned short values[MAX_DEVICES];
		int i;
		for (i = 0; i < MAX_DEVICES; i++)
			values[i] = 1; 
		sem_init.array = values; 
		semctl(semID_dev, 0, SETALL, sem_init);
	} else if (errno==EEXIST) {
		// the semaphore set already exists, so just use it
		if ((semID_dev = semget(SEM_KEY, 1, 0666)) < 0) {
			perror("semget devices");
		}
		PRINTDEBUG(DBG_SEMAPHORES, 
			   ("using existing device semaphores %d\n", 
			    semID_dev));
	} else {
		perror("semget global");
	}
}

static inline void grab_dev_semaphore(int index) {
	struct sembuf sem = {index, -1, SEM_UNDO};
	
	if (semop(semID_dev, &sem, 1) == -1) {
		perror("grab device semaphore");
	}
	PRINTDEBUG(DBG_SEMAPHORES, 
		   ("Grabbed device semaphore at index %d, value is %d\n", 
		    index, semctl(semID_dev, index, GETVAL, NULL)));
}

static inline void release_dev_semaphore(int index) {
	struct sembuf sem = {index, 1, SEM_UNDO };

	if (semop(semID_dev, &sem, 1) == -1) {
		perror("release semaphore");
	}
	PRINTDEBUG(DBG_SEMAPHORES, 
		   ("Released device semaphore at index %d, value is %d\n", 
		    index,
		    semctl(semID_dev, index, GETVAL, NULL)));
}

/** 
 * The library constructor assures that if libadwin is compiled
 * without KERNEL_MODULE, the semaphores are properly set up and the
 * the config file is read.
 */
void __attribute__ ((constructor)) my_init(void) {

#ifdef HAVE_SYS_UTSNAME_H
	struct utsname name;
        /*
         * We have <sys/utsname.h>, so we assume we have "uname()".
         */
        if (uname(&name) < 0) {
		strncpy(name.sysname, "uname failed", sizeof(name.sysname) - 1);
		strncpy(name.release, "uname failed", sizeof(name.release) - 1);
		strncpy(name.machine, "uname failed", sizeof(name.machine) - 1);
        }
#endif

	snprintf(VersionString, sizeof(VersionString) - 2, 
		 "%s\t%s\t%s\t%s\t%s\t%s\t%s", 
		 PACKAGE_STRING, __VERSION__, 
#ifdef HAVE_SYS_UTSNAME_H
		 name.sysname, name.release, name.machine,
#else
		 "unknown sysname", "unknown release", "unknown machine",
#endif

#ifdef ADWIN_BIG_ENDIAN
		 "Big Endian",
#else 
		 "Little Endian",
#endif

#ifdef __LP64__
		 "64-bit"
#else
		 "32-bit"
#endif
		 );

#ifdef KERNEL_MODULE
	return;
#else
	init_dev_semaphore();
	init_glob_semaphore();
	grab_glob_semaphore();
	reloadConfigFile();
	release_glob_semaphore();
#endif
}

/** 
 * SETMUTEX has two purposes: determining the device configuration
 * from the device_no and providing synchronization for device access.
 *
 * If libadwin was compiled with KERNEL_MODULE, then SETMUTEX reads
 * the device information for device "device" from
 * /proc/adwin/dev/0x%X with a ioctl. Synchsonization is implemented
 * by file locks.
 *
 * If libadwin was compiled without KERNEL_MODULE, then SETEMUTEX
 * reads the device information for device "device" from the config
 * file (actually, the file is read and parsed only once and then kept
 * in memory). Synchronization is implemented using semaphores.
 *
 * @param device The device number to lock and to get *vd for.
 * @param vd pointer to the the corresponting adwin_dev entry.
 * 
 * @return If libadwin was compiled with KERNEL_MODULE, a
 * filedescriptor to the appropriate device is returned, otherwise,
 * the index of the device in the device list is returned.
 *
 * In both cases -1 is returned on failure.
 */
inline int SETMUTEX(int device_no, adwin_dev_t *vd) {
	int dev_fd;

#ifdef KERNEL_MODULE
	char file[80];
	// open device
	sprintf(file, "/proc/adwin/dev/0x%X", device_no);
	if ((dev_fd = open(file, O_RDWR)) < 0) {
		// device not found
		if ((errno == ENOENT) || (errno == ENODEV))
			adwin_errno = k_this_link_is_unknown;
		
		return -1;
	}
	// load device info into vd
	ioctl(dev_fd, ADIOGKDEVINFO, vd);
#else
	// look for device in config file
	if ((dev_fd = findDevice(device_no)) == -1) {
		// device not found
		PRINTDEBUG(DBG_LIBADWIN, ("device %d not found\n", device_no));
		adwin_errno = k_this_link_is_unknown;
		// reset vd to defaults
		clearDevice(vd);
		return -1;
	}
	// grab semaphore for desired device
	grab_dev_semaphore(dev_fd);
	// load device info into vd
	*vd = getDevice(dev_fd);
#endif

	switch(vd->dev_type) {
#ifdef WITH_LINK
	case DC_LINK:
	case DC_PCMCIA:
		return dev_fd;
#endif
#ifdef WITH_USB
	case DC_USB:
		adwin_errno = -ENOTSUP;
		return -1;
#endif
#ifdef WITH_ETHERNET
	case DC_ETHERNET:
		if (vd->dev_processor != DP_DSP)
			vd->dev_processor = DP_DSP;
		// here we could check the software version of the device
		// instead of doing it when we need the info...
		// maybe todo.

		/*if (! vd->dev_server_version) {
		  if((vd->dev_server_version = __get_Version_TcpIp(vd)) < 0) {
		  adwin_errno = k_wrong_version;
		  return -1;
		  }
		  }*/
		return dev_fd;
#endif
	case DC_UNKNOWN:
	default:	
		adwin_errno = k_unknown_device;
		return -1;
	}
}

/** 
 * SETMUTEX_INIT_GEN_DRIVER is similar to SETMUTEX, with the exception
 * that it calls Set_DeviceNo to make sure that the generic driver is
 * properly initialized.
 * 
 * @param device_no see SETMUTEX
 * @param vd see SETMUTEX
 * 
 * @return see SETMUTEX
 */
inline int SETMUTEX_INIT_GEN_DRIVER(int device_no, adwin_dev_t *vd) {
	Set_DeviceNo(device_no);
	return SETMUTEX(device_no, vd);
}

/** 
 * CLEARMUTEX is the counterpart to SETMUTEX. Each time SETMUTEX is
 * successfully called for a device, CLEARMUTEX must be called
 * eventually to release the mutex.
 * 
 * @param vd Pointer to the currently used adwin_dev.
 * @param dev_fd Device number of the currently used adwin_dev.
 */
inline void CLEARMUTEX(adwin_dev_t *vd, int dev_fd) {

#ifndef KERNEL_MODULE
	// if we don't use the kernel module, we must only release the
	// semaphore.
	release_dev_semaphore(dev_fd);
	return;
#endif
	// if we use the kernel module, release the mutex by closing
	// the fd.
	switch(vd->dev_type) {
#ifdef WITH_LINK
	case DC_LINK:
	case DC_PCMCIA:
		close(dev_fd);
		break;
#endif
#ifdef WITH_USB
	case DC_USB:
		break;
#endif
#ifdef WITH_ETHERNET
	case DC_ETHERNET:
		close(dev_fd);
		break;
#endif
		default:
			return;
	}
}

/** 
 * Return the connection type for device_no.
 * 
 * @param device_no device_no to get connection type of.
 * 
 * @return connection type of device_no.
 */
int32_t Get_Connection_Type(short device_no) {
	adwin_dev_t dev;
	int dev_fd;

	if ((dev_fd = SETMUTEX (device_no, &dev)) >= 0) {
		CLEARMUTEX (&dev, dev_fd);
		switch(dev.dev_type) {
#ifdef WITH_ETHERNET
		case DC_ETHERNET:
			return 2;
#endif
#ifdef WITH_USB
		case DC_USB: // just a stub, there is currently no USB support.
			if (USB_connected(&dev)) 
				return 1;
			else 
				return 11;
#endif
#if defined(WITH_LINK)
		case DC_LINK:
		case DC_PCMCIA:
			return 0;
#endif
		default:
			return -1;
		}
	}
	return -1;
}

// note: buf must hold at least INET_ADDRSTRLEN (16) chars!
int Get_IP_Address(char* buf, short device_no) {
	adwin_dev_t dev;
	int dev_fd;

	if ((dev_fd = SETMUTEX (device_no, &dev)) >= 0) {
		CLEARMUTEX (&dev, dev_fd);
		inet_ntop(AF_INET, (const void*)&dev.dev_net_server.sin_addr, 
			  buf, INET_ADDRSTRLEN); // should not fail...
		return 0;
	}
	return -1;
}

int Get_Port(short device_no) {
	adwin_dev_t dev;
	int dev_fd;

	if ((dev_fd = SETMUTEX (device_no, &dev)) >= 0) {
		CLEARMUTEX (&dev, dev_fd);
		return ntohs(dev.dev_net_server.sin_port);
	}
	return -1;
}

int Get_Timeout(short device_no) {
	adwin_dev_t dev;
	int dev_fd;

	if ((dev_fd = SETMUTEX (device_no, &dev)) >= 0) {
		CLEARMUTEX (&dev, dev_fd);
		return dev.dev_net_timeout;
	}
	return -1;
}

void Get_Memory_dsp(int32_t base_addr, int32_t count, int32_t *dest) {
	adwin_get_memory_dsp(base_addr, count, dest, CURRENTDEVICE.DeviceNo);
}

int adwin_get_memory_dsp(int32_t base_addr, int32_t count, int32_t *dest, int id) {
	adwin_dev_t dev;
	int fd, tmp = 255;

	if ((fd = SETMUTEX_INIT_GEN_DRIVER(id, &dev)) < 0)
		return tmp;

/* 	SELECT_PROTOCOL_WITH_RETURN2(tmp, Get_Memory_dsp, &dev, fd, count, dest, base_addr); */
	SELECT_PROTOCOL(tmp, Get_Memory_dsp, &dev, fd, base_addr, count, dest);

	CLEARMUTEX(&dev, fd);
	return tmp == -1 ? 255 : 0;
}

/* int adwin_clear_data (int dataID, int id) { */
/* 	adwin_dev_t dev; */
/* 	int fd, tmp = 1; */

/* 	if ((fd = SETMUTEX_INIT_GEN_DRIVER(id, &dev)) < 0) */
/* 		return tmp; */

/* 	SELECT_PROTOCOL(tmp, Clear_Data, &dev, fd, dataID); */

/* 	CLEARMUTEX(&dev, fd); */
/* 	return tmp; */
/* } */

/*****************************************************************************/

int  adwin_get_data_type(int data_no, int id) {
	adwin_dev_t dev;
	int fd, tmp = 255;

	if ((fd = SETMUTEX_INIT_GEN_DRIVER(id, &dev)) < 0)
		return tmp;

	SELECT_PROTOCOL(tmp, GetDataType, &dev, fd, data_no);	// returns 0 on error

	if ((dev.dev_processor == DP_DSP) && (tmp == ADWIN_SYSTEM_SHORT))
		tmp = ADWIN_SYSTEM_LONG;

	//printf("adwin_system_type for data %d is %d\n", data_no, tmp);

	CLEARMUTEX(&dev, fd);
	return tmp;
}


char *libadwin_version() {
	return PACKAGE_VERSION;
}


#define DEBUG_VERSION 1
int32_t get_internal_version () {
	char *version = VERSION;
	char v[40];
	int vi = 0;
	int n;
	
	// copy all numeric chars from VERSION to v
	for (n = 0; version[n]; n++) 
		if (version[n] >= '0' && version[n] <= '9') 
			v[vi++] = version[n];
	// null-terminate v
	v[vi] = 0;
	
	return strtol (v, NULL, 10) | (DEBUG_VERSION << 24);
}

short Get_Data_packed (void *tarr, short typ, short nadc, int32_t lstart, int32_t il ,short nadr) {
	/*   typ = 1 für short
	     typ = 2 für long
	     typ = 5 für float
	     typ = 6 für double  */

     switch (typ)
     {
        case 1:  //short
	{
	    adwin_errno=k_function_not_allowed;
	    return 255;
	}
     	case 2:  //long
	{
	    adwin_errno=k_function_not_allowed;
	    return 255;
	}
	case 5:  //float
	{
	    adwin_errno=k_function_not_allowed;
	    return 255;
	}
	case 6:  //double
	{
	    adwin_errno=k_function_not_allowed;
	    return 255;
	}

	default :
	{
            adwin_errno=k_function_not_allowed;
	    return 255;
	}

     }
}

// WARNING: this function is not reentrant
const char *adwin_strerror(int e) {
	int i;

	if (e == 0)
		return "No Error.";

	if (e < 0)
		return strerror(-e);

	for (i = 0; adwin_errlist[i].__errstr; i++)
		if (adwin_errlist[i].__errno == e)
			return adwin_errlist[i].__errstr;

	return "No description available";
}

// reentrant version of adwin_strerror
int adwin_strerror_r(int e, char* buf, size_t n) {
	int i;

	if (e == 0) {
		if (strlen("No Error.") < n) {
			strcpy(buf, "No Error");
			return 0;
		} else {
			errno = ERANGE;
			return -1;
		}
	}

	if (e < 0) {

#ifdef STRERROR_R_CHAR_P
		strerror_r(-e, buf, n);
		return 0;
#else
		return strerror_r(-e, buf, n);
#endif
	}

	for (i = 0; adwin_errlist[i].__errstr; i++)
		if (adwin_errlist[i].__errno == e) {
			if (strlen(adwin_errlist[i].__errstr) < n) {
				strcpy(buf, adwin_errlist[i].__errstr);
				return 0;
			} else {
				errno = ERANGE;
				return -1;
			}
		}

	if (strlen("No description available") < n) {
		strcpy(buf, "No description available");
		return 0;
	} else {
		errno = ERANGE;
		return -1;
	}
}

void adwin_perror(const char *msg) {
	char buf[MAX_ERRSTR_LEN];
	adwin_strerror_r(adwin_errno, buf, MAX_ERRSTR_LEN);
	fprintf(stderr, "%s: %s\n", msg, buf);
}

adwin_errlist_t adwin_errlist[] = {
	{ k_timeout_to_link, 		"Timeout error on writing to the ADwin-system." },
	{ k_timeout_from_link,		"Timeout error on reading from the ADwin-system."},
	{ k_this_link_not_allowed,	"The device No. is not allowed."},
	{ k_this_link_is_unknown,	"The device No. is not known."},
	{ k_unknown_device,		"Device is unknown."},
	{ k_function_not_allowed,	"Function for this device not allowed." },
	{ k_wrong_version,		"Incompatible versions of ADwin operating system ,\ndriver (ADwin32.DLL) and/or ADbasic binary-file." },
	{ k_Data_is_to_small,		"The Data is too small." },
	{ k_Fifo_is_to_small,		"The Fifo is too small or not enough values." },
	{ k_Fifo_to_less_data,		"The Fifo has not enough values." },
	{ k_File_not_found,		"File not found." },
	{ k_File_not_binaer,		"The file is not an ADBasic binary-file." },
	{ k_File_not_valid,		"The file is not valid." },
	{ k_File_not_btl,		"The file is not a BTL." },
	{ k_WrongBinFile,		"ADbasic binary-file is for another processortype or damaged." },
	{ k_error_net_error,		"Network error (RPC)." },
	{ k_error_net_binding_compose,	"Network error (Bind)." },
	{ k_error_net_stringbinding,	"Network error (String bind)." },
	{ k_error_net_password,		"Wrong password (RPC)." },
	{ k_error_tcp_error,		"Network error (RPC)." },
	{ k_error_tcp_timout_read,	"Network timeout." },
	{ k_error_tcp_password,		"Wrong password." },
	{ 0, NULL},
};

int check_type_loc(const char *file, int line, adwin_data_type_t adt) {
	if ((adt == ADWIN_SHORT)
	    || (adt == ADWIN_LONG)
	    || (adt == ADWIN_FLOAT)
	    || (adt == ADWIN_DOUBLE)
	    || (adt == ADWIN_VARIANT))
		return 0;
#ifdef ADWIN_DEBUG
	printf("%s:%d:invalid adwin type %d found!\n", file, line, adt);
#endif
	return 1;
}

int check_system_type_loc(const char *file, int line, adwin_system_data_type_t asdt) {
	if ((asdt == ADWIN_SYSTEM_SHORT)
	    || (asdt == ADWIN_SYSTEM_INTEGER)
	    || (asdt == ADWIN_SYSTEM_LONG)
	    || (asdt == ADWIN_SYSTEM_FLOAT))
		return asdt;
#ifdef ADWIN_DEBUG
	printf("%s:%d:invalid adwin type %d found!\n", file, line, asdt);
#endif
	return asdt;
}

adwin_system_data_type_t to_system_type(adwin_data_type_t adt) {
	switch (adt) {
	case ADWIN_SHORT:
		return ADWIN_SYSTEM_SHORT;
	case ADWIN_LONG:
		return ADWIN_SYSTEM_LONG;
	case ADWIN_FLOAT:
		return ADWIN_SYSTEM_LONG; //todo?
	case ADWIN_DOUBLE:
	case ADWIN_VARIANT:
		// these two cannot be converted directly.
	default:
		;
#ifdef ADWIN_DEBUG
		printf("Invalid type for conversion encountered (%d)\n", adt);
#endif
	}
	return ADWIN_SYSTEM_LONG;
}

